
#include "error.h"
#include "identifier.h"
#include "primtype.h"
#include "treetype.h" 
#include "tree.h"
#include "data.h"
#include "flowgraph.h"
#include "program.h" 
#include "runtimestate.h"
#include "substitution.h"
#include "varstore.h"
#include "typesystem.h"
#include "parser.h"


#include "identifier.stat"

void testtokenizer( ) 
{

   tokenizer tt;

   tt. scan( );

   // Test the tokenizer:

   while( tt. lookahead. size( ) && 
          tt. lookahead. front( ). type != tkn_EOF )
   {
      std::cout << tt. lookahead. front( ) << "\n";
      tt. lookahead. pop_front( ); 
      tt. scan( ); 
   }

   // Actually if the size of lookahead gets 0, something went wrong.

   ASSERT( tt. lookahead. size( ));
   std::cout << tt. lookahead. front( ) << "\n";
      // This is an EOF token. 
}


tree readtree( varstore& vs ) throw( error ) 
{
   std::cout << "\n"; 
   std::cout << "Please type a tree to be typechecked: ";
   tokenizer tt; 
   parser( tt, vs, tkn_B1, 10 );

   if( tt. lookahead. size( ) != 2 ||
       tt. lookahead. front( ). type != tkn_B1 ||
       tt. lookahead. back( ). type != tkn_SEMICOLON )
   {
      error err;
      err. s << "stopped due to parsing error";
      throw err;
   }
   
   return tt. lookahead. front( ). tt. front( ); 
}


int main( int argc, char* argv [ ] )
{
   std::cout << "hallo\n";

   for( unsigned int i = 0; i < 1; ++ i )
   {
      primtype p1 = primtype( identifier( "int" ), 4 );
      primtype p2 = primtype( identifier( "double" ), 8 );

      treetype tt1 = treetype( type_prim, p1 );
      treetype tt2 = treetype( type_prim, p2 );

      std::list< treetype > args1;
      std::list< treetype > args2;

      args1. push_back( treetype( type_prim, p1 ));
      args1. push_back( treetype( type_void ));
      args1. push_back( tt1 );

      args2. push_back( treetype( type_prim, p1 ));
      args2. push_back( treetype( type_void ));

      std::cout << ( treetype( type_func, tt1, args1 ) == 
                     treetype( type_func, tt1, args1 ) ) << "\n";
 
      treetype g = treetype( type_func, tt2, args1 );
      std::cout << g. getargtype(2) << "\n";


      // Here starts code that you need in exercise 9. 

      // We build treetypes char/int/double:

      treetype tt_bool = treetype( type_prim, primtype( "bool", sizeof(bool)));
      treetype tt_char = treetype( type_prim, primtype( "char", sizeof(char))); 
      treetype tt_int = treetype( type_prim, primtype( "int", sizeof(int)));
      treetype tt_double = treetype( type_prim, 
                                     primtype( "double", sizeof( double )));

      tree pi = tree( "pi", 
                      tobinary( 3.141592653589793238462643383279 ),
                      tt_double );

      pi. prettyprint( std::cout );  
      
      tree e = tree( "e", 
                     tobinary( 2.718281828459045235360287471352 ),
                     tt_double );

      tree zero = tree( "zero", tobinary(0), tt_int );
      tree one = tree( "one", tobinary(1), tt_int );
      tree two = tree( "two", tobinary(2), tt_int );
      tree t = tree( "+", tree( "*", pi, zero ), two ); 
      t. prettyprint( std::cout );

      // We put a few function declarations in functionlist.
      // Can be used as example in task 2 and task 3: 

      typesystem ts; 
     
      std::list< treetype > argtypes;
      argtypes. push_back( tt_int );
      argtypes. push_back( tt_int );
 
      ts. instances [ identifier( "+" ) ]. push_back( 
              typesystem::concretefunc( identifier( "+int" ), 
              treetype( type_func, tt_int, argtypes )));

      argtypes. clear( );
      argtypes. push_back( tt_double );
      argtypes. push_back( tt_double );

      ts. instances [ identifier( "+" ) ]. push_back(
              typesystem::concretefunc( identifier( "+double" ),
              treetype( type_func, tt_double, argtypes )));

      argtypes. clear( );
      argtypes. push_back( treetype( type_pointer, 
                              treetype( type_prim, primtype( "X", 0 ))));
      argtypes. push_back( tt_int ); 

      ts. instances [ identifier( "+" ) ]. push_back(
              typesystem::concretefunc( identifier( "+pntr" ), 
                                        identifier( "X" ),
              treetype( type_func, treetype( type_pointer,
                              treetype( type_prim, primtype( "X", 0 ))),
                              argtypes )));
              
      std::cout << "Typesystem:\n\n";

      ts. convfunctions. push_back( typesystem::convfunc( "conv_int_double", 4,
                                              tt_int, tt_double )); 
      ts. convfunctions. push_back( typesystem::convfunc( "conv_char_int", 5,
                                              tt_char, tt_int )); 
      ts. convfunctions. push_back( typesystem::convfunc( "conv_char_double", 
                                              50,
                                              tt_char, tt_double ));
      ts. convfunctions. push_back( typesystem::convfunc( "array2pointer", 1,
                                              identifier( "X" ),
                                              treetype( type_ref, 
                                                 treetype( type_array, 0, 
                                                    treetype( type_prim, 
                                                       primtype( "X", 0 )))),
                                              treetype( type_pointer, 
                                                 treetype( type_prim,
                                                    primtype( "X", 0 )))));
      std::cout << ts << "\n";

      std::cout << ts. allconversions( 
                  tree( "x", tobinary(44), 
                    treetype( type_ref, 
                       treetype( type_array, tt_char ))));


      // If you want, you can assign more variables:

      varstore vs;

      vs. assign( "x", tt_double );
      vs. assign( "a", tt_double );
      vs. assign( "b", tt_double );
      vs. assign( "c", tt_double );

      vs. assign( "i", tt_int );
      vs. assign( "j", tt_int );
      vs. assign( "k", tt_int );

      vs. assign( "table", treetype( type_array, 10, tt_int ));
      vs. assign( "p", treetype( type_pointer, tt_double ));
      vs. assign( "q", treetype( type_pointer, tt_double ));
      vs. assign( "matrix", treetype( type_array, 5, 
                               treetype( type_array, 10, tt_double )));
      // Read one tree. 

      try
      {
         tree res = readtree( vs ); 
         std::cout << "read:\n";
         res. prettyprint( std::cout ); 
      } 
      catch( const error& err )
      {
         std::cout << err << "\n";
      }
      return 0;

      // We make our first program: 

      flowgraph fl;
      {
         flowgraph::iterator it = fl. begin( );
         instruction ins( instruction::move_pointer, 
                          instruction::argument( "x", tt2 ), 
                          instruction::argument( "x" ));

         fl. insert( ins, it, 0 );
         it = it. succ(0);

         ins = instruction( instruction::move_pointer,
                            instruction::argument( "y", tt2 ),
                            instruction::argument( "y" ));
         fl. insert( ins, it, 0 );
         it = it. succ(0);

         ins = instruction( instruction::move_double,
                            instruction::argument( tobinary( 33.0 )),
                            instruction::argument( "x", 0 ));

         fl. insert( ins, it, 0 );   
         it = it. succ(0);

         ins = instruction( instruction::move_double,
                            instruction::argument( tobinary( 12.0 )),
                            instruction::argument( "y", 0 )); 
   
         fl. insert( ins, it, 0 );
         it = it. succ(0);

         ins = instruction( instruction::mult_double,
                            instruction::argument( "x", 0 ),
                            instruction::argument( "y", 0 ),
                            instruction::argument( "z" )); 

         fl. insert( ins, it, 0 );
         it = it. succ(0); 

         ins = instruction( instruction::print_double,
                            instruction::argument( "z" ));
         fl. insert( ins, it, 0 );
         it = it. succ(0);

      }
   
      program pr;
      pr. insert( "main", treetype( type_void ), fl ); 
      pr. insert( "hallo", treetype( type_void ), fl ); 
      std::cout << pr << "\n";

      runtimestate rs( 256 );
      rs. tracing = true;
 
      try 
      {
         call( pr, "main", rs );
      }
      catch( const error& err )
      {
         std::cout << err << "\n";
      }
     
   }
   std::cout << "\n";
   identifier::printnormtable( std::cout );

}



